/* (C) Ensita.NET 2003,2004 */
/* (C) Softzone ltd 2004 */
/* $Id: clamfilter.c,v 1.3 2004/12/02 18:50:28 egor Exp $ */

#ifndef VERSION
#define VERSION "debug"
#endif

#include <pwd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>

#ifndef DELIVER
#define DELIVER "/usr/sbin/sendmail"
#endif

#ifndef CLAMDSCAN
#define CLAMDSCAN "/usr/bin/clamdscan"
#endif

#ifndef VIRUS_NOTIFY
#define VIRUS_NOTIFY "From: ClamAV notify <%s>\n" \
	"Subject: a virus found in a message to you\n" \
	"\n\n" \
	"Hello. A virus is found in a message addressed to you.\n\n" \
	"      From address: \"%s\"\n" \
	"      Subject: \"%s\"\n\n" \
	"      Message-id: \"%s\"\n" \
	"      Envelope from: %s\n" \
	"      Envelope to: %s\n" \
	"\n\n" \
	"-- \nYours, Clam antivirus scanner, version " VERSION "\n"
#endif


// from postfix's sys_exits.h
#define EX_OK           0
#define EX_USAGE        64              /* command line usage error */
#define EX_OSERR        71              /* system error (e.g., can't fork) */
#define EX_CANTCREAT    73              /* can't create (user) output file */
#define EX_IOERR        74              /* input/output error */
#define EX_TEMPFAIL     75              /* temporary failure */


char temporalfile[] = "/tmp/clamfilter.XXXXXX";

char *from_, *subject_, *messageid_;

int debug = 0;


//search for header content
int header_line(char *header, char *data, char **result) {
	char *tmp;
	//search for header start point in data
	if (*result = (char *)strcasestr(data,header)) // ! why strcasestr() not strstr() ?
	{
		//skip header, move to header content
		*result += strlen(header);
		//search for header content end
		if (!(tmp = strstr(*result,"\n")))
		{
			//not found - all header
			tmp = *result+strlen(*result);
		} 
	}
	if (!*result) {	//check header in first line of data
		header = header+1;
		//search for header start point in data
		if (*result = (char *)strcasestr(data,header)) // ! why strcasestr() not strstr() ?
		{
			//skip header, move to header content
			*result += strlen(header);
			//search for header content end
			if (!(tmp = strstr(*result,"\n")))
			{
				//not found - all header
				tmp = *result+strlen(*result);
			} 
		}
	}
	//return header content lenght
	return (*result)?(tmp-*result):0;
}


int read_message_to_file()
{
	int size;
	int incsize=0;
	int tmpfd = mkstemp(temporalfile);
	char buffer[102401];

	if (tmpfd == -1) {
		syslog(LOG_MAIL|LOG_ERR, "read_message_to_file: " \
			"could not create temporary file");
		exit(EX_CANTCREAT);
	}
	while( (size=read(0, buffer, 102400)) != 0) {
		if (size < 0) { 
			syslog(LOG_MAIL|LOG_ERR, 
				"read_message_to_file: " \
				"could not read next portion of data " \
				"from stdin");
			unlink(temporalfile);
			exit(EX_IOERR);
		}

		if (debug) {
			syslog(LOG_MAIL|LOG_DEBUG, "read_message_to_file: " \
				"read %d bytes", size);
		}
		
		incsize += size;
		
		if(write(tmpfd,buffer,size) < 0) { 
			syslog(LOG_MAIL|LOG_ERR, 
				"read_message_to_file: " \
				"could not write to temporal file");
			unlink(temporalfile);
			exit(EX_IOERR);
		}
	}
	
	if (debug) {
		syslog(LOG_MAIL|LOG_DEBUG, "read_message_to_file: totally " \
			"passed %d bytes", incsize);
	}
	return tmpfd;
}

int deliver_file(int fd, const char *returnpath, const char *receipent)
{
	pid_t child;
	int childread[2],parentread[2];
	int readfd,writefd;
	int status;
	char buffer[102400];
	int size;

	if (pipe(childread)) { 
		syslog(LOG_MAIL|LOG_ERR, "deliver_file: " \
			"could not pipe(childread)");
		return (EX_IOERR); 
	}
	if (pipe(parentread)) { 
		syslog(LOG_MAIL|LOG_ERR, "deliver_file: " \
			"could not pipe(parentread)");
		return (EX_IOERR); 
	}
	lseek(fd,0,SEEK_SET);
	child = fork();	
	if (!child) {
		// i'm a child
		readfd=childread[0]; close(childread[1]);
		writefd=parentread[1]; close(parentread[0]);
		close(0);
		close(1);
		close(2); 
	 	dup2(open("/dev/null","w"),2);
		dup2(readfd,0);
	 	dup2(open("/dev/null","w"),1);
		if (debug) { 
			syslog(LOG_MAIL|LOG_DEBUG, "deliver_file: " \
				"I run %s -i -f %s %s", DELIVER, returnpath, receipent);
		}
		if (execl(DELIVER,DELIVER, "-i", "-f",
			returnpath, receipent, NULL) == -1) { 
			syslog(LOG_MAIL|LOG_ERR, "deliver_file: " \
				"child: cannot exec or what? errno=%d", errno);
			return(EX_TEMPFAIL);
		}
	} else if (child>0) {
		// i'm a parent
		writefd=childread[1]; close(childread[0]);
		readfd=parentread[0]; close(parentread[1]);

		size = read(fd,buffer,102400);
		if (size == -1) { 
			syslog(LOG_MAIL|LOG_ERR, "deliver_file: " \
				"parent: cannot read buffer");
			return(EX_IOERR); 
		}
		while (size>0) {
			if (write(writefd,buffer,size) == -1) { 
				syslog(LOG_MAIL|LOG_ERR, 
					"deliver_file: " \
					"parent: cannot write buffer");
				return(EX_IOERR);
			}
			size = read(fd,buffer,102400);
			if (size == -1) { 
				syslog(LOG_MAIL|LOG_ERR, 
					"deliver_file: " \
					"parent: cannot read buffer");
				return(EX_IOERR);
			} 
		}
		close(writefd); 
		waitpid(child,&status,0);
		unlink(temporalfile);
		return (WEXITSTATUS(status));
	} else {
		// oops
		syslog(LOG_MAIL|LOG_ERR, "deliver_file: could not " \
			"fork or what?");
		return (EX_OSERR);
	}
	if (debug) {
		syslog(LOG_MAIL|LOG_DEBUG, "deliver_file: success");
	}
}

int deliver_warning(int fd, const char *returnpath, const char *receipent, const char *fromemail)
{
	pid_t child;
	int childread[2],parentread[2];
	int readfd,writefd;
	int status;
	char buffer[102400];
	int size;

	if (pipe(childread)) { 
		syslog(LOG_MAIL|LOG_ERR, "deliver_warning: " \
			"could not pipe(childread)");
		return (EX_IOERR); 
	}
	if (pipe(parentread)) { 
		syslog(LOG_MAIL|LOG_ERR, "deliver_warning: " \
			"could not pipe(parentread)");
		return (EX_IOERR); 
	}
	child = fork();	
	if (!child) {
		// i'm a child
		readfd=childread[0]; close(childread[1]);
		writefd=parentread[1]; close(parentread[0]);
		close(0);
		close(1);
		close(2); 
	 	dup2(open("/dev/null","w"),2);
		dup2(readfd,0);
	 	dup2(open("/dev/null","w"),1);
		if (execl(DELIVER,DELIVER, "-i", "-f",
			fromemail, receipent, NULL) == -1) { 
			syslog(LOG_MAIL|LOG_ERR, "deliver_warning: " \
				"child: cannot exec or what? errno=%d", errno);
			return(EX_TEMPFAIL);
		}
	} else if (child>0) {
		FILE *templatefile;
		char buffer[204800];
		
		// i'm a parent
		writefd=childread[1]; close(childread[0]);
		readfd=parentread[0]; close(parentread[1]);

		sprintf(buffer, VIRUS_NOTIFY, fromemail, from_, 
			subject_, messageid_, returnpath, receipent);
		
		if (write(writefd,buffer,strlen(buffer)) == -1) { 
			syslog(LOG_MAIL|LOG_ERR, 
				"deliver_warning: " \
				"parent: cannot write buffer");
			return(EX_IOERR);
		}
		close(writefd); 
		waitpid(child,&status,0);
		unlink(temporalfile);
		return (WEXITSTATUS(status));
	} else {
		// oops
		syslog(LOG_MAIL|LOG_ERR, "deliver_warning: could not " \
			"fork or what?");
		return (EX_OSERR);
	}
	if (debug) {
		syslog(LOG_MAIL|LOG_DEBUG, "deliver_warning: success");
	}
}

int check_file()
{
	pid_t child;
	int childread[2],parentread[2];
	int readfd,writefd;
	int status;
	char buffer[102401];
	int size;

	if (pipe(childread)) { 
		syslog(LOG_MAIL|LOG_ERR, "check_file: " \
			"could not pipe(childread)");
		return (EX_IOERR); 
	}
	if (pipe(parentread)) { 
		syslog(LOG_MAIL|LOG_ERR, "check_file: " \
			"could not pipe(parentread)");
		return (EX_IOERR); 
	}
	child = fork();	
	if (!child) {
		// i'm a child
               	readfd=childread[0]; close(childread[1]);
		writefd=parentread[1]; close(parentread[0]);
		close(0);
		close(1);
		close(2);
		dup2(open("/dev/null","w"),2);
		dup2(readfd,0);
		dup2(open("/dev/null","w"),1);
		int res=execl(CLAMDSCAN, CLAMDSCAN, "--quiet", temporalfile, NULL);
		syslog(LOG_MAIL|LOG_ERR, "check_file: " \
			"child: cannot exec, return code %d", res);
		return(EX_TEMPFAIL);
		
	} else if (child>0) {
		// i'm a parent
		waitpid(child,&status,0);
		if (debug) {
			syslog(LOG_MAIL|LOG_DEBUG, "check_file: waitpid " \
				"finished and returned %d", WEXITSTATUS(status));
		}
		return (WEXITSTATUS(status));
	} else {
		// oops
		syslog(LOG_MAIL|LOG_ERR, "check_file: could not " \
			"fork or what?");
		return (EX_OSERR);
	}
	if (debug) {
		syslog(LOG_MAIL|LOG_DEBUG, "check_file: success");
	}
}


void read_args(int argc, char** argv, const char** returnpath, 
		const char** receipent, const char ** fromemail) {
	int opt, i, j; int restlen;

	while(-1 != (opt = getopt(argc, argv, "f:F:D"))) {
		switch(opt) {
		case 'D':
			debug=1;
			break;
		case 'f':
			*returnpath = optarg;
			break;
		case 'F':
			*fromemail = optarg;
			break;
		}
	}

	*receipent = malloc(65535);
	restlen=65534;
	strcpy(*receipent, "");
	for (i=optind;i<argc;i++) { 
		strncat(*receipent, argv[i], restlen);
		strncat(*receipent, " ", restlen);
		restlen-=strlen(argv[i])+1;
		if (restlen<=0) { 
			syslog(LOG_MAIL|LOG_ERR, "read_args: too long receipent list!");
			exit (EX_IOERR);
		}
	}
	return;
}

void parse_content(int inputfd) {
	int len; int size;
	char buffer[204800];
	char *pointer;

	lseek(inputfd,0,SEEK_SET);
	size = read(inputfd,buffer,204800);
	if (size < 0) { 
		syslog(LOG_MAIL|LOG_ERR, 
			"parse_content: " \
			"could not read first 204800 bytes of data from" \
			"file %s", temporalfile);
		
		from_ = (char *)strdup("(unknown from)");
		subject_ = (char *)strdup("(unknown subject)");
		messageid_ = (char *)strdup("(unknown messageid)");
		return;
	}

	len = header_line("\nfrom: ", buffer, &pointer);
	if (len) { 
		from_ = malloc(len+2);
		strncpy(from_, pointer, len);
		from_[len]=0;
	} else { 
		from_ = (char *)strdup("(unknown from)");
	}

	len = header_line("\nsubject: ", buffer, &pointer);
	if (len) { 
		subject_ = malloc(len+2);
		strncpy(subject_, pointer, len);
		subject_[len]=0;
	} else { 
		subject_ = (char *)strdup("(unknown subject)");
	}

	len = header_line("\nmessage-id: ", buffer, &pointer);
	if (len) { 
		messageid_ = malloc(len+2);
		strncpy(messageid_, pointer, len);
		messageid_[len]=0;
	} else { 
		messageid_ = (char *)strdup("(unknown messageid)");
	}

	if (debug) {
		syslog(LOG_MAIL|LOG_DEBUG, 
			"deliver_warning: " \
			"message from = '%s', subject = '%s', msgid = '%s'", 
			from_, subject_, messageid_);
	}
}


int main (int argc, char **argv) {
	int inputfd = -1;
	int outputfd = -1;
	int exitcode = 0;
	int len = 0;
	int virusresult = 0;
	const char *returnpath = NULL;
	const char *receipent = NULL;
	const char *fromemail = NULL;

	read_args(argc, argv, &returnpath, &receipent, &fromemail);

	openlog("clamfilter", LOG_PID, LOG_MAIL);

	if(returnpath==NULL || receipent==NULL || fromemail==NULL) {
		printf("This program is intended to be called by postfix, " \
			"not directly.\n");
		syslog(LOG_MAIL|LOG_ERR, "Wrong usage, exit");
		exit(EX_USAGE);
	}

      	inputfd = read_message_to_file();

	virusresult = check_file();

	if(virusresult == 1) {
		/* nothing its virus */
		parse_content(inputfd); 

		syslog(LOG_MAIL|LOG_DEBUG, "message is " \
			"virus, warning delivered");
		deliver_warning(inputfd, returnpath, receipent, fromemail); 
	} else if (virusresult == 0) { 
		exitcode=deliver_file(inputfd, returnpath, receipent);
		if (debug) {
			syslog(LOG_MAIL|LOG_DEBUG, "message is not " \
			"virus, delivered with exit code %d, return=" \
			"%s, receipent=%s", exitcode, 
			returnpath, receipent);
		} else { 
			syslog(LOG_MAIL|LOG_ERR, "message is clean" );
		}
	} else if (virusresult == 2) { 
		exitcode=deliver_file(inputfd, returnpath, receipent);
		syslog(LOG_MAIL|LOG_ERR, "could not connect " \
		"to clamd, so message is treated as not infected...");
	} else {
		syslog(LOG_MAIL|LOG_ERR, "message check " \
			"failed, return code=%d", virusresult); 
		exit(virusresult);
	}

	unlink(temporalfile);
	if (debug) {
		syslog(LOG_MAIL|LOG_DEBUG, "removed temp file %s, exit with " \
		"code=%d", temporalfile, exitcode);
	}

	return exitcode;
}
